"""
This helperclass reads TWISS files in CSV format produced by MADX.
See more information in twissreader.py.
"""
__author__ = "Christian Staufenbiel"
__date__ = "03/09/2019"
__version__ = "0.1"

import numpy as np
import pandas as pd

class TwissReaderCSV():
    """ Stores the header and data of a TWISS file."""

    # Initialize
    def __init__(self, filename, sep=','):
        """ Initialize object
        Parameters
        ----------
        filename : string
            Path to twiss file
        sep : string
            Seperator used in CSV file
        """
        # Class variables
        self.header = {}
        self.data = pd.DataFrame([])
        # Open file
        s = self.__readfile(filename)
        split = self.__split_header_data(s, sep=sep)

    def __readfile(self, filename):
        """Reads the TwissFile

        Parameters
        ----------
        filename : string
            Path to file
        Returns
        -------
        string
            Content of TwissFile
        """
        # Try to open CSV file
        try:
            f = open(filename, 'r')
        except FileNotFoundError:
            print('TwissReader: The file {} was not found!'.format(filename))
            raise
        # Read string from file and return
        res = f.read()
        f.close()
        return res

    def __split_header_data(self, s, sep):
        """ Splits string of a twiss file to header and data.
        We assume that the header has always 2 entries (variable name, value)
        and the data has more than 2 entries.

        Parameters
        ----------
        s : string
            string of the twiss file (generated by __readfile())
        sep : string
            character used as seperator in the twiss file

        Returns
        -------
        list (strings)
            Containing seperated lines of twiss file

        """
        # Split on line break
        split = s.split('\n')
        # Get the number of entries (noe) per line
        # This will be used to determine between header and data
        noe_list = []
        for l in split:
            noe = len(l.split(sep))
            noe_list.append(noe)
        noe_list = np.array(noe_list)
        # Check if there is only header and data
        # Note: The line before the data, containing descriptions, has one column
        #       less than the data lines, since data lines end with a comma
        #       Also there is one last row in the file, which only has one entry
        #       Hence there are 4 different numbers of entries in a valid TwissFile
        if np.unique(noe_list).shape[0] > 4:
            raise RuntimeError('TwissReader: More than 2 sections in file! ')

        # Get the unique number of elements per line, sorted by occurrence
        # I.e. number of elements in header first, then datadescription, then data, then last row with only one entry
        unique, index = np.unique(noe_list, return_index=True)
        unique = unique[index.argsort()]

        # Get indices where noe per line are the same
        idx_set = [np.argwhere(noe_list == x) for x in unique]
        header_idx = idx_set[0].flatten()  # Indices of header
        descr_data_idx = idx_set[1].flatten()  # Indices of variable description
        data_idx = idx_set[2].flatten()  # Indices of data
        footer_idx = idx_set[3].flatten()  # Indices of footer

        # Create header dictionary
        self.header = self.__create_dictionary(split, header_idx, sep=sep)

        # Create pandas dataframe
        self.data = self.__create_dataframe(split, descr_data_idx, data_idx, sep=sep)

        # Returning split of data
        return split

    def __create_dictionary(self, split, idx, sep):
        """Creates a dictionary from the split values and given indices

        Parameters
        ----------
        split : list (strings)
            String of twiss file split by each row
        idx : list (integers)
            Indices corresponding to the header
        sep : string
            Seperator used to for name,value seperation

        Returns
        -------
        dictionary
            Dictionary with all header information

        """
        dict = {}
        # Iterate through all indices corresponding to the dictionary
        for i in idx:
            # Split into variable name and value
            l = split[i].split(sep)
            # Raise error if there is more than one name and value
            if len(l) > 2:
                raise RuntimeError('TwissReader: Can not form dict!')
            key, val = l
            # Remove the " characters from the names and variables
            key = key.replace('"', '')
            val = val.replace('"', '')
            # Try to convert value to float
            try:
                val = float(val)
            except:
                pass
            # Add key/val to dictionary
            dict[key] = val

        return dict

    def __create_dataframe(self, split, descr_idx, data_idx, sep):
        """ Creates a dataframe given the index for description and data.

        Parameters
        ----------
        split : list (strings)
            String of twiss file split by each row
        descr_idx : list (integers)
            Index which row holds the description
        data_idx : list (integers)
            Indices which rows hold the data
        sep : string
            Seperator that is used to separate name and value

        Returns
        -------
        pandas.DataFrame
            DataFrame containing all the data stored in TwissFile
        """
        # Retrieve column names
        if len(descr_idx) > 1:
            raise RuntimeError('Too many indices for description!')
        # Replace " by no char to convert have clean names without " around it
        col_names = split[descr_idx[0]].replace('"', '')
        col_names = col_names.split(sep)

        # Retrieve data
        data = []
        for i in data_idx:
            # Remove " to convert to float later on
            l = split[i].replace('"', '')
            # Remove ending comma (TODO: CLEANER WAY)
            l = l[:-2]
            # Split line and convert to float if possible
            vals = l.split(sep)
            for i, val in enumerate(vals):
                val = val.replace(' ', '')  # remove whitespaces
                try:
                    vals[i] = float(val)
                except:
                    continue
            # Append to data array
            data.append(vals)

        # Create DataFrame and return
        df = pd.DataFrame(data, columns=col_names)
        df = df.set_index(col_names[0])
        return df
